#                              -*- Mode: Perl -*- 
#!/app/unido-i06/magic/perl
#                              -*- Mode: Perl -*- 
use Config;
use File::Basename qw(&basename &dirname);

# List explicitly here the variables you want Configure to
# generate.  Metaconfig only looks for shell variables, so you
# have to mention them as if they were shell variables, not
# %Config entries.  Thus you write
#  $startperl
# to ensure Configure will look for $Config{startperl}.

# This forces PL files to create target in same directory as PL file.
# This is so that make depend always knows where to find PL derivatives.
chdir(dirname($0));
($file = basename($0)) =~ s/\.PL$//;
$file =~ s/\.pl$//
	if ($^O eq 'VMS' or $^O eq 'os2');  # "case-forgiving"

open OUT,">$file" or die "Can't create $file: $!";

print "Extracting $file (with variable substitutions)\n";

# In this section, perl variables will be expanded during extraction.
# You can use $Config{...} to use Configure variables.

print OUT <<"!GROK!THIS!";
$Config{'startperl'} 
!GROK!THIS!
print OUT <<'!NO!SUBS!';
# iman -- interactive manual search
# ITIID           : $Source: Man-Index/iman_ $ $Revision: 1.9 $
# Author          : Ulrich Pfeifer
# Created On      : Fri 25 Apr 14:19:19 1997
# Last Modified By: Ulrich Pfeifer
# Last Modified On: Sun Apr 27 12:45:57 1997
# Language        : Perl
# Status          : Demo for TPJ
# 
# (C) Copyright 1997, Ulrich Pfeifer, all rights reserved.
# 
!NO!SUBS!

# In the following, perl variables are not expanded during extraction.

print OUT <<'!NO!SUBS!';
eval 'exec perl  -S $0 "$@"'
    if 0;

use Man::Index;
use strict;
use Getopt::Long;
use Term::ReadKey;
use vars qw(%OPT $VERSION);
use Term::Cap;

# $Format: "$VERSION = \"$ProjectVersion$\";"$ 
$VERSION = "0.19";

print "iman - interactive manual search $VERSION by Ulrich Pfeifer\n";

%OPT = (
        'index' => '/tmp/man',
       );

GetOptions(\%OPT,
           'index=s',
          ) or die "USAGE: $0 [-index <path>]\n";

my $INDEX = new Man::Index 'index' => $OPT{'index'};
die "Could not open index '$OPT{'index'}'\n" unless $INDEX;

my $TERMINAL = eval {Tgetent Term::Cap { TERM => undef, OSPEED => 9600 }};
sub keypress () {
  print STDERR "Press any key";
  ReadKey 0;
  print $TERMINAL->Tputs('cl') if $TERMINAL; 
}

# cbreak mode
ReadMode 3;
END { ReadMode 0};

# get terminal size
my ($WIDTH, $HEIGHT) = GetTerminalSize;
$HEIGHT = 24 unless $HEIGHT; 
$WIDTH  = 80 unless $WIDTH;

my $CURSOR;
my $QUERY;
my @HIT;
my %FUNC =
  (
   'q' => sub { exit },
   's' => \&search,
   'l' => \&menu,
   'p' => sub { $CURSOR--},
   'n' => sub { $CURSOR++},
   'r' => sub { $HIT[$CURSOR]->{mark} = '+'; $CURSOR++}, 
   'i' => sub { $HIT[$CURSOR]->{mark} = '-'; $CURSOR++}, 
   'x' => sub { $HIT[$CURSOR]->{mark} = ' '; $CURSOR++},
   'd' => sub { printf ("%5d %s\n",
                        $HIT[$CURSOR]->{docno},
                        $HIT[$CURSOR]->{path}); keypress },
   'm' => \&run_man,
   'f' => \&feedback,
   'h' => \&help,
  );
@FUNC{'?', '+', '-', ' ', "\101", "\102"} = @FUNC{'h', 'r', 'i', 'n', 'p', 'n'};
@FUNC{"\033", "\133"} = (sub{}) x 2;

while (1) {
  print STDERR "\r> ";
  my $key  = ReadKey 0;
  my $func = $FUNC{$key} || \&help;
  &$func;
  $CURSOR = 0 if $CURSOR < 0;
  $CURSOR = $#HIT if $CURSOR > $#HIT;
  print $TERMINAL->Tputs('ho') if $TERMINAL; 
  &menu;
}

sub help {
  # $Format: "  print \"$ProjectHeader$\\n\";"$
  print "tpj 0.19 Sun, 27 Apr 1997 13:36:16 +0200 pfeifer\n";
  print <<EOH
>
  p     select previous entry
  n     select next entry
  r,+   mark as relevant
  i,-   mark as nonrelevant
  x     clear mark
  s     simple search
  f     feedback search
  q     exit the program
  m     run 'man -l' on the page
  f     run relevance feedback query
  l     redisplay hit list
  d     display path
?
  enter text
EOH
  ;
  keypress;
}

sub menu {
  my $cursor;
  print "\r";
  for (@HIT) {
    my $cm   =  (($cursor == $CURSOR)?'=':' ');
    $HIT[$cursor]->{line} ||=
      substr(sprintf ("%5.3f ",$_->{score})
             . $_->{headline} . ' ' x $WIDTH, 0, $WIDTH-5);
    print $cm, '[',  $_->{mark}, '] ', $HIT[$cursor++]->{line}, "\n";
  }
  
}

sub search {
  print STDERR "\r? ";
  ReadMode 1;
  $QUERY = ReadLine;
  ReadMode 3;

  my $score = $INDEX->search($QUERY);
  unless ($score) {print "No hits\n"; return;}

  @HIT      = ();               # reset state
  $CURSOR   = 0;
  my $display = $HEIGHT-1;

  for my $doc_no (sort {$score->{$b} <=> $score->{$a}} keys %$score) {
    push @HIT, {
                docno    => $doc_no,
                score    => $score->{$doc_no},
                headline => $INDEX->headline($doc_no),
                path     => $INDEX->path($doc_no),
                mark     => ' ',
               };
    last unless --$display;
  }
  1;
}

sub feedback {
  my %seen;
  my @word = grep !$seen{$_}++, $INDEX->words($QUERY);
  unless (@word) {
    print STDERR "\rNo query given!\n";
    keypress; return;
  }
  my (%good, %all);
  for (@HIT) {
    if ($_->{mark} eq '+') {
      $good{$_->{docno}} = 1;
    }
    if ($_->{mark} ne ' ') {
      $all{$_->{docno}} = 1;
    } 
  }
  my $good = scalar keys %good;
  my $all  = scalar keys %all;
  unless (scalar %all) {
    print STDERR "\rNo relevance judgments!\n";
    keypress; return;
  }
  my %score;
  for my $word (@word) {
    my ($f, $r);
    my %document = $INDEX->documents($word);
    for my $docno (keys %all) {
      if ($document{$docno}) {       # document contains the word
        $f++;
        $r++ if $good{$docno};
      }
    }
    printf "\r%-10s in %2d/%2d judged/relevant documents\n", $word, $f, $r;
    my $p = ($r+0.5)/($good+1);
    my $q = ($f-$r+0.5)/($all-$good+1);
    for my $docno (keys %document) {
      $score{$docno} += log($p*(1-$q)/$q/(1-$p));
    }
  }

  @HIT      = ();               # reset state
  $CURSOR   = 0;
  my $display = $HEIGHT-1;

  for my $doc_no (sort {$score{$b} <=> $score{$a}} keys %score) {
    my $mark = ' ';
    $mark = '-' if $all{$doc_no};
    $mark = '+' if $good{$doc_no};
    push @HIT, {
                docno    => $doc_no,
                score    => $score{$doc_no},
                headline => $INDEX->headline($doc_no),
                path     => $INDEX->path($doc_no),
                mark     => $mark,
                };
    last unless --$display;
  }
  keypress;
}

sub run_man {
  unless ($HIT[$CURSOR]) {
    print STDERR "\rNo manual selected\n";
    keypress; return;
  }
  my $path = $HIT[$CURSOR]->{path};
  unless (-f $path) {
    print STDERR "\rNo manual $path disapeared\n";
    keypress; return;
  }
  
  if (system 'man', '-l', $path) {
    warn "Could not run 'man -l $path: $?\n";
    keypress;
  }
  ReadMode 3;
}

!NO!SUBS!

close OUT or die "Can't close $file: $!";
chmod 0755, $file or die "Can't reset permissions for $file: $!\n";
exec("$Config{'eunicefix'} $file") if $Config{'eunicefix'} ne ':';
